/*************************************************************************
ID: popa3d/virtual.c

Portions Copyright 2002, Bulent Yilmaz

Part of Microbrew Microbrew Message Center
Multi-server Mailer Suite

http://www.microbrew.org/products/mmc/

Microbrew Message Center is free software; you can redistribute it and/or  
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

Microbrew Message Center is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License
along with MicroSieve; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

A full copy of the GNU General Public License can be found at:
http://www.gnu.org/licenses/gpl.txt
**************************************************************************/

/*
 * Virtual domain support.
 */

#include "params.h"

#if POP_VIRTUAL

#define _XOPEN_SOURCE 4
#define _XOPEN_SOURCE_EXTENDED
#define _XOPEN_VERSION 4
#define _XPG4_2
#define __USE_SVID
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef NAME_MAX
#define NAME_MAX			255
#endif

extern int log_error(char *s);
extern struct passwd *fgetpwent (FILE *);

char *virtual_domain;
char *virtual_spool;

int virtual_startup(void)
{
	return 0;
}

static int is_valid_user(char *user)
{
	unsigned char *p;

/* This is pretty liberal, but we're going to use direct syscalls only,
 * and they have to accept all the printable characters */
	for (p = (unsigned char *)user; *p; p++)
		if (*p < ' ' || *p > 0x7E || *p == '/') return 0;

	if (p - (unsigned char *)user > NAME_MAX) return 0;

	return 1;
}

struct passwd *virtual_userpass(char *user, char *pass, int *known)
{
	struct passwd *pw;
	struct stat stat;
	char *address, *pathname, *tempuser;
	FILE *passfile;
	int match=0;
	char *myuser;
	
	/* Start out with a good paranoid set of defaults. */
	
	*known = 0;
        virtual_domain = NULL;
        
	log_error("virtual_userpass");

        /* Check to make sure that ther are no malicious characters in the 
         * supplied username. */

	if (!is_valid_user(user)) {
		log_error("Invalid user");
		user = "INVALID";
		return NULL;
	}

	/* Copy username data since strtok() will destroy my data as a free
	 * gift.  Pull out the username@domain format.  */

	if((tempuser = strdup(user))) {
		myuser = strtok(tempuser, "@");
		address = strtok(NULL, "@");
	} else {
		log_error("Input username strdup");
		user = "INVALID";
		return NULL;
	}
//      user = "INVALID";

        /* Make sure that both the username and the domain are valid. */

	if ((!address)||(!myuser)) {
		log_error("Invalid username format");
		free(tempuser);
		return NULL;
	}
	
        /* Make copy of the data I want and free the memory.  Be pedantic 
         * about it. */ 

	if( !(myuser=strdup(myuser)) ) {
		free(tempuser);
		return NULL;
	}	
	if( !(address=strdup(address)) ) {
		free(myuser);
		free(tempuser);
		return NULL;
	}
	free(tempuser);
	

	/* This "can't happen", but is just too critical to not check 
	 * explicitly. */

	
	if (strchr(address, '/') || strchr(myuser, '/')) {
		free(myuser);
		free(address);
		return NULL;
	}

        /* Allocate memory and determine my path to the virtual domain. */

	pathname = malloc(strlen(VIRTUAL_HOME_PATH) + strlen(address) + strlen(VIRTUAL_AUTH_PATH) + strlen(myuser) + 4);
	if (!pathname) {
		log_error("Pathname malloc()");
		free(myuser);
		free(address);
		return NULL;
	}
	sprintf(pathname, "%s/%s", VIRTUAL_HOME_PATH, address);

        /* Check to make sure that my path to the virtual domain actually 
         * exists. */

	if (lstat(pathname, &stat)) {
		log_error("lstat");
		free(myuser);
		free(address);
		free(pathname);
		return NULL;
	}

        /* Determine my path to the auth file for this user. */

	sprintf(pathname, "%s/%s/%s/%s", VIRTUAL_HOME_PATH, address, VIRTUAL_AUTH_PATH, myuser);

	/* Open up the file and see if we can get a '/etc/passwd' style entry. 
	 * for the specified user. */

	if ( (passfile = fopen(pathname, "r")) ) {

		match = 1;
		do {
			pw = fgetpwent(passfile);
			if( pw == NULL) break;
			
		} while ( (( match = strcmp(pw->pw_name, myuser) )!=0) && (!feof(passfile)) );

		fclose(passfile);
		free(pathname);
		free(myuser);
		
		if(match) {
			log_error("Missing auth data");
			free(address);
			return NULL;
		}
			
	} else {	
		log_error("Auth file open");
		free(address);
		free(myuser);
		free(pathname);
		return NULL;
	}
	
	/* It exists.  Proceed to set my virtual_domain to that address.  
         * At this point, address shouldn't be freed, and hopefully it will
         * get done later. */

	virtual_domain = address;

        /* Compare the supplied password with the password from the file. */

	if(!strcmp( crypt(pass, pw->pw_passwd), pw->pw_passwd )) {

		/* The user authenticates, lets point virtual_spool at the 
		 * user's mailbox.  Again, be pedantic. */

		if( (virtual_spool = malloc(strlen(VIRTUAL_HOME_PATH) + strlen(virtual_domain) + strlen(VIRTUAL_SPOOL_PATH) + 3)) == NULL) {
			log_error("virtual_spool malloc");
			free(address);
			virtual_domain = NULL;
			return NULL;
		}	
		sprintf(virtual_spool, "%s/%s/%s", VIRTUAL_HOME_PATH, virtual_domain, VIRTUAL_SPOOL_PATH);

		/* Wipe passwords from memory. */
		memset(pw->pw_passwd, 0, sizeof(pw->pw_passwd));
		
		/* Set the mailbox name (user) correctly. */
		strncpy(user, pw->pw_name, sizeof(user));
		*(user+strlen(pw->pw_name))=0;
		
		/* Return the data. */
		*known = 1;		
		return pw;		
	} else {
		log_error("password mismatch");
		free(address);
		free(myuser);
		virtual_domain = NULL;
		return NULL;
	}
	
	/* Just in case: */
	
	return NULL;
	
}

#endif
